/// Defines a rust wrapper trait to match the Dpki exposed function trait definition
/// for the conductor to use in the various dpki related flows, e.g. bootstrap and agent addition
use crate::holochain::Holochain;
use holochain_core_types::{dna::capabilities::CapabilityRequest, error::HolochainError};
use holochain_json_api::json::JsonString;
use holochain_persistence_api::cas::content::AddressableContent;

use holochain_core::{
    context::Context, nucleus::actions::call_zome_function::make_cap_request_for_call,
};

pub const DPKI_ZOME_NAME: &str = "dpki";
pub const DPKI_TRAIT_FN_ADD_AGENT: &str = "create_agent_key";
pub const DPKI_TRAIT_FN_INIT: &str = "init_dpki";
pub const DPKI_TRAIT_FN_IS_INITIALIZED: &str = "is_initialized";

use std::{convert::TryInto, sync::Arc};

pub trait DpkiInstance {
    fn dpki_create_agent_key(&mut self, agent_name: String) -> Result<(), HolochainError>;
    fn dpki_init(&mut self, params: String) -> Result<(), HolochainError>;
    fn dpki_is_initialized(&mut self) -> Result<bool, HolochainError>;
}

/// create a capability request for a given dpki call
fn dpki_cap_request(
    context: Arc<Context>,
    function: &str,
    parameters: &str,
) -> Result<CapabilityRequest, HolochainError> {
    let token = context.agent_id.address();
    Ok(make_cap_request_for_call(
        context,
        token,
        function,
        JsonString::from_json(parameters),
    ))
}

#[holochain_tracing_macros::newrelic_autotrace(HOLOCHAIN_CONDUCTOR_LIB)]
impl DpkiInstance for Holochain {
    /// wrapper for the dpki create_agent_key trait function
    fn dpki_create_agent_key(&mut self, agent_name: String) -> Result<(), HolochainError> {
        let params = json!({ "agent_name": agent_name }).to_string();
        let cap_request = dpki_cap_request(self.context()?, DPKI_TRAIT_FN_ADD_AGENT, &params)?;
        let _result = Holochain::call_zome_function(
            self.context()?,
            DPKI_ZOME_NAME,
            cap_request,
            DPKI_TRAIT_FN_ADD_AGENT,
            &params,
        )?;
        Ok(())
    }

    // wrapper for the dpki init trait function
    fn dpki_init(&mut self, params: String) -> Result<(), HolochainError> {
        let params = json!({ "params": params }).to_string();
        let cap_request = dpki_cap_request(self.context()?, DPKI_TRAIT_FN_INIT, &params)?;
        let _result = Holochain::call_zome_function(
            self.context()?,
            DPKI_ZOME_NAME,
            cap_request,
            DPKI_TRAIT_FN_INIT,
            &params,
        )?;
        Ok(())
    }

    // wrapper for the dpki is_initialized trait function
    fn dpki_is_initialized(&mut self) -> Result<bool, HolochainError> {
        let params = "{}";
        let cap_request = dpki_cap_request(self.context()?, DPKI_TRAIT_FN_IS_INITIALIZED, params)?;
        let result = Holochain::call_zome_function(
            self.context()?,
            DPKI_ZOME_NAME,
            cap_request,
            DPKI_TRAIT_FN_IS_INITIALIZED,
            params,
        )?;
        let result: Result<bool, HolochainError> = result.try_into()?;
        match result.to_owned() {
            Ok(r) => println!("DPKI instance is initialized: {:?}", r),
            Err(e) => println!("ERROR: DPKI is not initialized: {:?}", e),
        }
        result
    }
}
